from typing import List

from aiomisc import Service, entrypoint
from aiomisc.service.raven import RavenSender
from aiomisc_log import basic_config

from yatracker_linker.args import Parser
from yatracker_linker.deps import config_deps
from yatracker_linker.service import HttpService


def main():
    parser = Parser(
        auto_env_var_prefix='YATRACKER_LINKER_',
        config_files=[
            '.yatracker-linker.ini',
            '~/.yatracker-linker.ini',
            '/etc/yatracker-linker.ini'
        ],
    )
    parser.parse_args()
    basic_config(level=parser.log_level, log_format=parser.log_format)

    config_deps(parser)

    services: List[Service] = [
        HttpService(address=parser.address, port=parser.port)
    ]

    if parser.sentry.dsn:
        services.append(
            RavenSender(
                sentry_dsn=parser.sentry.dsn,
                client_options={'environment': parser.sentry.env}
            )
        )

    with entrypoint(
        *services,
        log_level=parser.log_level,
        log_format=parser.log_format
    ) as loop:
        loop.run_forever()


if __name__ == '__main__':
    main()



from http import HTTPStatus

from aiohttp import ClientSession
from yarl import URL


class StClient:
    def __init__(
        self,
        session: ClientSession,
        url: URL,
        token: str,
        link_origin: str
    ):
        self._session = session
        self._headers = {'Authorization': f'OAuth {token}'}
        self._base_url = url
        self._link_origin = link_origin

    def get_url(self, url_path: str) -> URL:
        return self._base_url / url_path.lstrip('/')

    async def issue_exists(self, key: str):
        url = self.get_url(f'issues/{key}')
        async with self._session.head(url, headers=self._headers) as resp:
            return resp.status == HTTPStatus.OK

    async def link_issue(self, key: str, mr_path: str):
        url = self.get_url(f'issues/{key}/remotelinks')
        json = {
            'origin': self._link_origin,
            'relationship': 'relates',
            'key': mr_path
        }
        async with self._session.post(
            url, headers=self._headers, json=json
        ) as resp:
            return resp.ok



from aiohttp import web
from aiomisc.service.aiohttp import AIOHTTPService

from yatracker_linker.st_client import StClient
from yatracker_linker.views.event import GitlabView


class HttpService(AIOHTTPService):
    __dependencies__ = (
        'st_client',
    )

    st_client: StClient

    async def create_application(self):
        app = web.Application()
        app.router.add_route('GET', GitlabView.URL_PATH, GitlabView)

        app['st_client'] = self.st_client

        return app



from aiohttp import ClientSession
from aiomisc_dependency import dependency, reset_store

from yatracker_linker.args import Parser
from yatracker_linker.st_client import StClient


async def st_client(parser: Parser):
    async with ClientSession() as session:
        return StClient(
            session=session,
            url=parser.tracker.url,
            token=parser.tracker.token,
            link_origin=parser.tracker.link_origin
        )


def config_deps(args):

    @dependency
    def parser() -> Parser:
        return args

    dependency(st_client)


def reset_deps():
    reset_store()



import asyncio
import logging
import re
from http import HTTPStatus
from typing import Any, List, Mapping

from aiohttp.web import HTTPBadRequest, Response, View

from yatracker_linker.st_client import StClient


PATTERN = re.compile(r'(?P<ticket>[a-z0-9]+-[0-9]+)', flags=re.IGNORECASE)

log = logging.getLogger(__name__)


def get_ticket_candidates(*items: str) -> List[str]:
    candidates = set()
    for item in items:
        if matches := PATTERN.findall(item):
            candidates.update(matches)

    return list(sorted(candidate.upper() for candidate in candidates))


def get_mr_url_path(url, project_path_with_namespace):
    index = url.find(project_path_with_namespace)
    return url[index:]


class GitlabView(View):
    URL_PATH = '/gitlab'

    @property
    def st_client(self) -> StClient:
        return self.request.app['st_client']

    async def get_tickets(self, event_data: Mapping[str, Any]) -> List[str]:
        candidates = get_ticket_candidates(
            event_data['object_attributes']['source_branch'],
            event_data['object_attributes']['target_branch'],
            event_data['object_attributes']['title'],
            event_data['object_attributes']['description']
        )

        existing = await asyncio.gather(*[
            self.st_client.issue_exists(candidate)
            for candidate in candidates
        ])

        return [
            candidate
            for candidate, exists in zip(candidates, existing)
            if exists
        ]

    async def post(self):
        event = await self.request.json()

        if event['event_type'] != 'merge_request':
            return Response(status=HTTPStatus.NO_CONTENT)

        mr_path = get_mr_url_path(
            event['object_attributes']['url'],
            event['project']['path_with_namespace']
        )
        if not mr_path:
            raise HTTPBadRequest(text='Unable to get merge request path')

        tickets = await self.get_tickets(event)
        if not tickets:
            log.info('No tickets to link with mr %s', mr_path)
            return Response(status=HTTPStatus.NO_CONTENT)

        log.info('Linking mr %s with tickets %r', mr_path, tickets)
        await asyncio.gather(*[
            self.st_client.link_issue(ticket, mr_path)
            for ticket in tickets
        ])

        return Response()



from typing import Optional

import argclass
from aiomisc_log import LogFormat
from yarl import URL


class SentryGroup(argclass.Group):
    dsn: Optional[URL]
    env: Optional[str]


class TrackerGroup(argclass.Group):
    url: URL
    token: str
    link_origin: str


class Parser(argclass.Parser):
    log_level: int = argclass.LogLevel
    log_format: str = argclass.Argument(
        choices=LogFormat.choices(),
        default=LogFormat.default()
    )
    address: str = argclass.Argument(default='0.0.0.0')
    port: int

    sentry = SentryGroup(title='Sentry options')
    tracker = TrackerGroup(title='Tracker options')
